package es.ucm.fdi.grasia.faerie.observer.internal.remote;

import java.net.URI;
import java.net.UnknownHostException;
import java.util.Arrays;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Set;

import org.apache.felix.scr.annotations.Component;
import org.apache.felix.scr.annotations.Properties;
import org.apache.felix.scr.annotations.Property;
import org.apache.felix.scr.annotations.Reference;
import org.apache.felix.scr.annotations.ReferenceCardinality;
import org.apache.felix.scr.annotations.ReferencePolicy;
import org.apache.log4j.Logger;

import es.ucm.fdi.grasia.faerie.core.AbstractAttributeUpdaterComponent;
import es.ucm.fdi.grasia.faerie.core.Attribute;
import es.ucm.fdi.grasia.faerie.core.Context;
import es.ucm.fdi.grasia.faerie.core.ContextElement;
import es.ucm.fdi.grasia.faerie.core.ContextEvent;
import es.ucm.fdi.grasia.faerie.core.ContextObserver;
import es.ucm.fdi.grasia.faerie.core.util.ContextURIs;

/**
 * Implementation of {@code ContextObserver} that observes the changes on remote elements and updates the local copies.
 *
 * @author Chema
 * @see Context
 */
@Component(inherit = true)
@Properties({ @Property(name = "service.remote.registration", boolValue = true),
		@Property(name = "service.exported.interfaces", value = "*") })
@Reference(name = "remoteContextObserver", referenceInterface = ContextObserver.class,
		cardinality = ReferenceCardinality.OPTIONAL_MULTIPLE, policy = ReferencePolicy.DYNAMIC,
		target = "(|(service.uri=*)(service.imported=*))")
@SuppressWarnings({ "unchecked", "rawtypes" })
public class RemoteObserver extends AbstractAttributeUpdaterComponent implements ContextObserver {

	private Map<URI, ContextObserver> remoteContextObservers = setUpRemoteContextObservers();
	private Map<URI, Set<URI>> elementsRequestingObservers = setUpElementsRequestingObservers();
	private Map<URI, Set<URI>> observersRequestedElements = setUpObserversRequestedElements();

	Map<URI, ContextObserver> setUpRemoteContextObservers() {
		return new HashMap<URI, ContextObserver>();
	}

	Map<URI, Set<URI>> setUpElementsRequestingObservers() {
		return new HashMap<URI, Set<URI>>();
	}

	Map<URI, Set<URI>> setUpObserversRequestedElements() {
		return new HashMap<URI, Set<URI>>();
	}

	ContextEvent createAddedEvent(ContextElement arg) {
		return new ContextEvent(this.getBundleEntity(), ContextEvent.ELEMENT_ADDED, arg);
	}

	ContextEvent createRemovedEvent(ContextElement arg) {
		return new ContextEvent(this.getBundleEntity(), ContextEvent.ELEMENT_REMOVED, arg);
	}

	ContextEvent createUpdateEvent(ContextElement source) {
		return new ContextEvent(source, ContextEvent.VALUE_UPDATED, this.getBundleEntity());
	}

	Set<URI> setUpIdSet2() {
		return new HashSet<URI>();
	}

	private void bindRequestingObserver(ContextElement element, ContextElement observer) {
		URI elementId = element.getId();
		URI observerId = ContextURIs.authorityURI(observer.getId());
		Set<URI> elementRequestingObservers = elementsRequestingObservers.get(elementId);
		if (elementRequestingObservers == null)
			elementsRequestingObservers.put(elementId, elementRequestingObservers = setUpIdSet2());
		elementRequestingObservers.add(observerId);
		Set<URI> observerRequestedElements = observersRequestedElements.get(observerId);
		if (observerRequestedElements == null)
			observersRequestedElements.put(observerId, observerRequestedElements = setUpIdSet2());
		observerRequestedElements.add(elementId);
	}

	private void unbindRequestingObserver(ContextElement element, ContextElement observer) {
		URI elementId = element.getId();
		URI observerId = ContextURIs.authorityURI(observer.getId());
		Set<URI> elementRequestingObservers = elementsRequestingObservers.get(elementId);
		if (elementRequestingObservers != null)
			elementRequestingObservers.remove(observerId);
		Set<URI> observerRequestedElements = observersRequestedElements.get(observerId);
		if (observerRequestedElements != null)
			observerRequestedElements.remove(elementId);
	}

	Map<URI, Attribute> setUpRemoteAttributeInstances() {
		return new HashMap<URI, Attribute>();
	}

	private Map<URI, Attribute> findRemoteAttributeInstances(URI remoteObsUri) {
		Map<URI, Attribute> result = setUpRemoteAttributeInstances();
		for (Entry<URI, Attribute> entry : this.getAllUpdatedAttributeInstances().entrySet())
			if (entry.getKey().getAuthority().equals(remoteObsUri.getAuthority()))
				result.put(entry.getKey(), entry.getValue());
		return result;
	}

	private Map<URI, ContextObserver> findRemoteObservers(URI remoteElemUri) {
		Map<URI, ContextObserver> result = setUpRemoteContextObservers();
		for (Entry<URI, ContextObserver> entry : remoteContextObservers.entrySet())
			if (entry.getKey().getAuthority().equals(remoteElemUri.getAuthority()))
				result.put(entry.getKey(), entry.getValue());
		return result;
	}

	private String propertiesToString(Map<String, Object> properties) {
		StringBuilder builder = new StringBuilder();
		for (Entry<String, Object> entry : properties.entrySet()) {
			builder.append("\n\t").append(entry.getKey()).append(" = ");
			Object value = entry.getValue();
			if (value.getClass().isArray())
				builder.append(Arrays.toString((Object[]) value));
			else
				builder.append(value);
		}
		return builder.toString();
	}

	private URI obtainAuthorityURI(Map<String, Object> properties) throws UnknownHostException {
		String uriStr = (String) properties.get("ecf.service.imported.endpoint.id");
		if (uriStr == null)
			uriStr = (String) properties.get("service.uri");
		URI uri = URI.create(uriStr);

		String host = ContextURIs.toCanonicalHostName(uri.getHost());

		return URI.create("//" + host + ":" + uri.getPort());
	}

	public void bindRemoteContextObserver(ContextObserver contextObserver, Map<String, Object> properties) {
		try {
			Logger.getLogger(this.getClass()).info(
					"Adding Remote " + ContextObserver.class.getSimpleName() + ": " + contextObserver
							+ propertiesToString(properties));

			URI authorityURI = obtainAuthorityURI(properties);

			remoteContextObservers.put(authorityURI, contextObserver);
			Logger.getLogger(this.getClass()).debug("Binding: " + authorityURI);

			Map<URI, Attribute> remoteAtts = findRemoteAttributeInstances(authorityURI);
			for (ContextElement remoteAtt : remoteAtts.values()) {
				Logger.getLogger(this.getClass()).debug("Remote Request: " + remoteAtt.getId() + " -> " + authorityURI);
				contextObserver.elementAdded(createAddedEvent(remoteAtt));
			}
			Set<URI> observerRequestedElements = observersRequestedElements.get(authorityURI);
			if (observerRequestedElements != null)
				for (URI observerRequestedElement : observerRequestedElements) {
					Logger.getLogger(this.getClass()).debug(
							"Remote Update: " + observerRequestedElement + " -> " + authorityURI);
					contextObserver
							.elementUpdated(createUpdateEvent(this.getObservedElement(observerRequestedElement)));
				}
		} catch (UnknownHostException ex) {
			Logger.getLogger(this.getClass()).error(ex.getLocalizedMessage(), ex);
		}
	}

	public void unbindRemoteContextObserver(ContextObserver contextObserver, Map<String, Object> properties) {
		try {
			Logger.getLogger(this.getClass()).info(
					"Removing Remote " + ContextObserver.class.getSimpleName() + ": " + contextObserver
							+ propertiesToString(properties));

			URI authorityURI = obtainAuthorityURI(properties);

			remoteContextObservers.remove(authorityURI);
		} catch (UnknownHostException ex) {
			Logger.getLogger(this.getClass()).error(ex.getLocalizedMessage(), ex);
		}
	}

	@Override
	public int handle(ContextEvent ce) {
		if (super.handle(ce) != ContextObserver.HANDLE_NONE)
			return super.handle(ce);
		Attribute elem = ce.getAttributeInstanceArg();
		return elem != null && !elem.getId().getAuthority().equals(ContextURIs.localURI().getAuthority())
				&& elem.isRemotelyUpdateable() ? ContextObserver.HANDLE_READ_WRITE : ContextObserver.HANDLE_NONE;
	}

	@Override
	public void elementAdded(Attribute att, ContextEvent event) {
		Logger.getLogger(this.getClass()).trace("(" + att + ", " + event + ") * " + this);
		Logger.getLogger(this.getClass()).trace("localURI() = " + ContextURIs.localURI());
		if (!att.getId().getAuthority().equals(ContextURIs.localURI().getAuthority())) {
			this.putUpdatedAttributeInstance(att);
			Map<URI, ContextObserver> remoteContObs = findRemoteObservers(att.getId());
			for (Entry<URI, ContextObserver> entry : remoteContObs.entrySet()) {
				Logger.getLogger(this.getClass()).debug(
						"Remote Request: " + att.getId() + " -> " + entry.getKey().getAuthority());
				entry.getValue().elementAdded(createAddedEvent(att));
			}
		} else {
			Logger.getLogger(this.getClass()).debug(
					"Remote Request: " + att.getId() + " <- " + event.getSource().getId().getAuthority());
			bindRequestingObserver(att, event.getSource());
			Attribute localAtt = this.contextRequest(att);
			this.putObservedElement(localAtt);
			elementUpdated(createUpdateEvent(localAtt));
		}
	}

	@Override
	public void elementRemoved(Attribute att, ContextEvent event) {
		Logger.getLogger(this.getClass()).trace("(" + att + ", " + event + ") * " + this);
		Logger.getLogger(this.getClass()).trace("localURI() = " + ContextURIs.localURI());
		if (!att.getId().getAuthority().equals(ContextURIs.localURI().getAuthority())) {
			Map<URI, ContextObserver> remoteContObs = findRemoteObservers(att.getId());
			for (Entry<URI, ContextObserver> entry : remoteContObs.entrySet()) {
				Logger.getLogger(this.getClass()).debug(
						"Remote Release: " + att.getId() + " -> " + entry.getKey().getAuthority());
				entry.getValue().elementRemoved(createRemovedEvent(att));
			}
			this.removeUpdatedAttributeInstance(att);
		} else {
			Logger.getLogger(this.getClass()).debug(
					"Remote Release: " + att.getId() + " <- " + event.getSource().getId().getAuthority());
			this.contextRelease(att.getId());
			this.removeObservedElement(att.getId());
			unbindRequestingObserver(att, event.getSource());
		}
	}

	@Override
	public void elementUpdated(Attribute att, ContextEvent event) {
		Logger.getLogger(this.getClass()).trace("(" + att + ", " + event + ") * " + this);
		Logger.getLogger(this.getClass()).trace("localURI() = " + ContextURIs.localURI());
		if (!att.getId().getAuthority().equals(ContextURIs.localURI().getAuthority())) {
			Logger.getLogger(this.getClass()).debug("Remote Update: " + att + " <- " + att.getId().getAuthority());
			Attribute localAtt = this.contextPeek(att.getId());
			if (localAtt != null)
				localAtt.update(att.getValue());
		} else {
			Set<URI> elementRequestingObservers = elementsRequestingObservers.get(att.getId());
			if (elementRequestingObservers != null)
				for (URI elementRequestingObserver : elementRequestingObservers) {
					Logger.getLogger(this.getClass()).debug(
							"Remote Update: " + att + " -> " + elementRequestingObserver.getAuthority());
					ContextObserver obs = remoteContextObservers.get(elementRequestingObserver);
					if (obs != null)
						obs.elementUpdated(event);
				}
		}
	}
}
